#include <fstream>
#include "EnvironMap.h"
#include "GlobalVars.h"
#include "Ray.h"
#include "Pane.h"
#include "rgbe.h"
#include "Halton.h"
#include "DataSet.h"
#include "FileUtils.h"

EnvironMap::EnvironMap(){
  gv = NULL;
  data = NULL;
  data3 = NULL;
  pweights = NULL;
  cweights = NULL;
  samples1 = samples2 = NULL;
  sint = NULL;
  Lavg = SbColor(0.0, 0.0, 0.0);
  worldToTextureRot.makeIdentity();
  textureToWorldRot.makeIdentity();
  dim[0] = dim[1] = 0;
  dim2[0] = dim2[1] = 0;
  dimll[0] = dimll[1] = 0;
  samples = 0;
  Lavg.setValue(0,0,0);
  worldToTextureRot.makeIdentity();
  textureToWorldRot.makeIdentity();
}

EnvironMap::~EnvironMap(){
  delete [] data;
  delete [] data3;
  
  for (int i=0; i<dimll[0]; i++)
    delete [] pweights[i];
  delete [] pweights;
  delete [] cweights;
  delete [] samples1;
  delete [] samples2;
  delete [] sint;
}

void EnvironMap::init(GlobalVars *gv_, float *inputPixels,
                      int pixX, int pixY, SbRotation *rot){

  // set GlobalVars pointer
  gv = gv_;

  if (rot) {
    setRotateSky(*rot);
  } else {
    // set environment map rotation
    // this works by rotating the incoming ray backwards
    // when doing a texture lookup
    float angle = gv->envMapRotation*M_PI/180.0; // convert degrees to radians
    setRotateSky(SbRotation(SbVec3f(0.0, 1.0, 0.0), angle));
  }

  float *image;
  bool imageDealloc = true;
  
  // use single color
  if (inputPixels) {
    dim[0] = pixX;
    dim[1] = pixY;
    image = inputPixels;
    imageDealloc = false;
    data = new SbColor[dim[X]*dim[Y]];
  } 
  else if (gv->environMapFile == "-1"){
    dim[0] = dim[1] = 1;
    image = new float[3 * dim[X]*dim[Y]];
    data = new SbColor[dim[X]*dim[Y]];
    image[0] = gv->worldBG[0];
    image[1] = gv->worldBG[1];
    image[2] = gv->worldBG[2];
  }
  // read HDR image
  else {
  
    // open file
    FILE *file = fopen(gv->environMapFile.c_str(), "rb");
    if (!file){

      printf("Could not open file %s.\n", gv->environMapFile.c_str());
      delete [] data;
      data = NULL;
      exit(1);
    }

    //
    //    0  ...  i (dim[0])
    //    .
    //    .      index(i,j) = dim[0]*j+i;
    //    .
    //    j
    // (dim[1])
    //    

    RGBE_ReadHeader(file, &dim[X], &dim[Y], NULL);

    // allocate space
    image = new float[3*dim[X]*dim[Y]];
    data = new SbColor[dim[0]*dim[1]];
  
    RGBE_ReadPixels_RLE(file, image, dim[X], dim[Y]);
    fclose(file);
  }
  
  // copy into memory
  int badVals = 0;
  for (int j=0; j<dim[Y]; j++)
    for (int i=0; i<dim[X]; i++){
      float *pixel = &image[(dim[0]*j+i)*3];

      // check for bad values
      for (int k=0; k<3; k++)
	if (!finite(pixel[k])){
	  pixel[k] = 0.0;
	  badVals++;
	}

      SbColor c(pixel[0], pixel[1], pixel[2]);
      //c *= 1.0/100.0;           // convert to hectowatts!!
      data[dim[0]*(dim[Y]-1-j)+i] = c;   
    }

  if (badVals)
    cerr << "Fixed " << badVals << " bad values in environment map "
	 << gv->environMapFile << ".\n";

  if (imageDealloc)
    delete [] image;

  // compute pixel weights
  cerr << "Computing environment map pixel weights\n";

  // downsample
  SbColor *data2;
  dim2[0] = MAX2((dim[0]+1)/2, 1);
  dim2[1] = MAX2((dim[1]+1)/2, 1);
  data2 = new SbColor[dim2[0]*dim2[1]];
  for (int j=0; j<dim2[1]; j++)
    for (int i=0; i<dim2[0]; i++){

      int x0, x1, y0, y1;
      x0=i*2;
      x1=x0+1;
      if (x1>=dim[0]) x1 = x0;
      y0=j*2;
      y1=y0+1;
      if (y1>=dim[1]) y1 = y0;

      // box filter
      SbColor c(0.0, 0.0, 0.0);
      c += data[y0*dim[0]+x0];
      c += data[y0*dim[0]+x1];
      c += data[y1*dim[0]+x0];
      c += data[y1*dim[0]+x1];
      data2[j*dim2[0]+i] = c/4.0;
    }
	   
  // temporarily swap downsampled texture
  // while calling getRadiance()
  int dimt[2];
  dimt[0] = dim[0];
  dimt[1] = dim[1];
  SbColor *datat = data;
  data = data2;
  dim[0] = dim2[0];
  dim[1] = dim2[1];

  // create lat long map
  dimll[0] = dim2[0];
  dimll[1] = MAX2(dim2[0]/2, 1);
  data3 = new SbColor[dimll[0]*dimll[1]];
  
  pweights = new float*[dimll[0]];
  for (int i=0; i<dimll[0]; i++)
    pweights[i] = new float[dimll[1]];
  cweights = new float[dimll[0]];

  // precompute sin(theta)
  sint = new float[dimll[1]];
  for (int j=0; j<dimll[1]; j++)
    sint[j] = sin((0.5+j)/dimll[1]*M_PI);

  // loop over lat/longs
  SbVec3f dir;
  float rsum = 0.0;
  Lavg = SbColor(0.0, 0.0, 0.0);
  float Lavg_weights = 0.0;
  
  for (int i=0; i<dimll[0]; i++){
    
    float csum = 0.0;
    for (int j=0; j<dimll[1]; j++){

      float theta, phi;
      theta = (0.5+j)/dimll[1]*M_PI;
      phi = (0.5+i)/dimll[0]*2.0*M_PI;
      phi -= M_PI_2;
      if (phi<0.0)
	phi += M_PI*2.0;
      dir.setValue(sin(theta)*cos(phi),
		   cos(theta),
		   sin(theta)*sin(phi));
      SbColor rad = getRadiance(dir);
      float b = brightness(rad);
      pweights[i][j]=b*sint[j];
      csum += pweights[i][j];

      // debugging
      b = clamp(b/4.0f, 0.0f, 1.0f);
      data3[j*dimll[0]+i] = SbColor(b,b,b);

      // average radiance
      Lavg += rad*sint[j];   // weighted by sint (smaller at poles) since there
                             // is too much of that counted in lat/long map
      Lavg_weights += sint[j];
    }

    cweights[i] = csum;
    rsum += csum;

    // normalize pixel weights
    if (csum){
      for (int j=0; j<dimll[1]; j++)
        pweights[i][j] /= csum;
    }
  }

  
  Lavg /= Lavg_weights;

  // reporting
  //cerr << "Lavg = " << Lavg << "\n";
  
  if ((Lavg[0]+Lavg[1]+Lavg[2])>0.00001)
    gv->sceneHasLights = 1;

  // normalize column weights
  if (rsum){
    for (int i=0; i<dimll[0]; i++)
      cweights[i] /= rsum;
  }

  // switch back texture
  dim[0]=dimt[0];
  dim[1]=dimt[1];
  data = datat;
  delete [] data2;

  // precompute samples
  
  // stratified sampling
  int sqrt_samples = (int)sqrt((float)gv->lightSamples*2);
  samples = sqrt_samples*sqrt_samples;
  samples1 = new double[samples];
  samples2 = new double[samples];

  // hammersley
  for (int i=0; i<samples; i++){
    samples1[i] = (double)i/samples;
    samples2[i] = RI_LP(i);
  }

//   for (int sampx=0; sampx<sqrt_samples; sampx++){
//     for (int sampy=0; sampy<sqrt_samples; sampy++){

//       // jittered
//       samples1[sampy*sqrt_samples+sampx]
// 	= (sampx + r.ctx->mydrand48()) / sqrt_samples;
//       samples2[sampy*sqrt_samples+sampx]
// 	= (sampy + r.ctx->mydrand48()) / sqrt_samples;
//     }
//   }
  
  return;
}


void EnvironMap::writeData3(){
  // spit out image of weights
  ofstream quick;
  quick.open("map.ppm");
  quick << "P3\n";
  quick << dimll[0] << " "
	<< dimll[1] << "\n";
  quick << "255\n";

  // find max value
  float max = 0.0;
  for (int j=0; j<dimll[1]; j++)
    for (int i=0; i<dimll[0]; i++){
      float m;
      m = cweights[i] * pweights[i][j];
      //m = data3[dimll[0]*j+i][1];
      if (m>max)
 	max = m;
    }
  
  for (int j=0; j<dimll[1]; j++)
    for (int i=0; i<dimll[0]; i++){
      int r = (int)(data3[dimll[0]*j+i][0]*255.0);
      int g = (int)(data3[dimll[0]*j+i][1]*255.0);
      int b = (int)(data3[dimll[0]*j+i][2]*255.0);

      float kf = cweights[i] * pweights[i][j] / max * 25.0 * 255.0;
      int k = (int)(kf);
      r = g = b = k; 
      
      quick << clamp(r, 0, 255) << " ";
      quick << clamp(g, 0, 255) << " ";
      quick << clamp(b, 0, 255) << " ";
      quick << "\n";
    }

  quick.close();
  exit(1);
}



SbColor EnvironMap::impSample(Ray &r, SbVec3f &n, SbVec3f &ng, int simple){

  /////////////////////////////////
  // importance sample the lights
  /////////////////////////////////

  SbColor color(0.0, 0.0, 0.0);

  // stratified sampling
  int sqrt_samples = (int)sqrt((float)gv->lightSamples*2);
  if (sqrt_samples<=0)
    return color;
  
  // use one sample in final gather
  if (simple)
   sqrt_samples = 1;

  int totalSamples = sqrt_samples*sqrt_samples;

  // deterministic sampling setup
  int ldDim1 = r.ldDimensionThis++;
  int ldDim2 = r.ldDimensionThis++;
  int bd1 = Halton::primes[ldDim1];
  int bd2 = Halton::primes[ldDim2];
  vector<unsigned short> &sigma1 = Halton::sigmas[ldDim1];
  vector<unsigned short> &sigma2 = Halton::sigmas[ldDim2];
  
  for (int sampx=0; sampx<sqrt_samples; sampx++){
    for (int sampy=0; sampy<sqrt_samples; sampy++){

      // jittered
      double eps1, eps2;
      if (gv->envLockSamples){
	eps1 = samples1[sampy*sqrt_samples+sampx];
	eps2 = samples2[sampy*sqrt_samples+sampx];
      } if (gv->quasiRandom) {

        // get two samples from ray sample tree
        // from Strictly Deterministic Sampling pg. 20
        Halton halt1, halt2, halt3;
        int j = sampx*sqrt_samples+sampy;
        halt1.Number(r.ldInstance, bd1, &sigma1[0]);
        halt2.Number(r.ldInstance, bd2, &sigma2[0]);
        halt3.Number(j, 2);
        eps1 = fmod(halt1 + j/float(totalSamples), 1);
        eps2 = fmod(halt2 + halt3, 1);
      }
      else {
        eps1 = (sampx + r.ctx->mydrand48()) / sqrt_samples;
        eps2 = (sampy + r.ctx->mydrand48()) / sqrt_samples;
      }

      // pick column
      double sum = 0, sumNext = 0;
      int c, lastGood = 0;
      for (c=0; c<dimll[0]-1; c++){
	if (cweights[c]<=0) continue;
	lastGood = c;
	sumNext += cweights[c];
	if (eps1<=sumNext) break;
	sum = sumNext;
      }
      if (cweights[c]<=0.) c = lastGood;
      eps1 = (eps1 - sum) / cweights[c];

      // pick row
      sum = sumNext = 0.0;
      lastGood = 0;
      int p;
      for (p=0; p<dimll[1]-1; p++){
	if (pweights[c][p] <= 0.) continue;
	lastGood = p;
	sumNext += pweights[c][p];
	if (eps2<=sumNext) break;
	sum = sumNext;
      }
      if (pweights[c][p] <= 0.) p = lastGood;
      eps2 = (eps2 - sum) / pweights[c][p];

      float phi = (c+eps1)/dimll[0]*2.0*M_PI;
      phi -= M_PI_2;
      if (phi<0.0)
	phi += M_PI*2.0;
      
      float theta = (p+eps2)/dimll[1]*M_PI;

      SbVec3f l;
      l.setValue(sin(theta)*cos(phi),
		 cos(theta),
		 sin(theta)*sin(phi));
      
      // dont illuminate lights below horizon
      if (ng.dot(l) <= 0.0)
	continue;

      // don't illuminate area that think they are facing away
      float dot1 = n.dot(l);
      if (dot1<=0.0)
	continue;      

      // trace from point to light point
      // i add a small amount to make it not intersect overlapping surfaces
      // i really hate doing this
      // optimize: fix this vv^^
      // fix: could solve with min_t
      const float t_eps = 0.002;
      Ray lr = Ray(r.x+l*t_eps, l, r.ctx);
      lr.ignore(r.prim);
      if (gv->pane->scenePrim->intersectRay(lr)) {
	continue; // hits something (and not background)
      }

      SbVec3f Le = getRadiance(l);
      float pw = cweights[c] * pweights[c][p] * dimll[0] * dimll[1] /
	(2.0*M_PI * M_PI * sint[p]);
      float value = dot1 / pw;

      // ray marching absorbtion
      if (gv->stepSize>0){
	float tau = 0.0;
	float sdt = gv->stepSize;
	float sigma_t = 0.0;
	SbVec3f lx = r.x+gv->pane->getMaxDist()*l;
	tau = gv->pane->dataSet->bresenSum(r.x, lx);
	tau = gv->meters(tau);
	value *= exp(-tau);
      }
	  
      // compute color
      // shirley p.120 eq. 15.5
      color += SbColor(r.diffCol[R] * M_1_PI * Le[R] * value,
		       r.diffCol[G] * M_1_PI * Le[G] * value,
		       r.diffCol[B] * M_1_PI * Le[B] * value) ;

    } // end stratified sampling y
  } // end stratified sampling x

  color /= totalSamples;
      
  // debugging
  //writeData3();
  
  return(color);
}

float EnvironMap::getSample(double eps1, double eps2,
			    SbVec3f &dir, SbColor &col){
  
      // pick column
      float sum = 0.0;
      int c;
      for (c=0; c<dimll[0]-1; c++){
	if (eps1<(sum+cweights[c]))
	  break;
	sum += cweights[c];
      }
      //eps1 = (eps1 - sum) / cweights[c];

      // pick row
      sum = 0.0;
      int p;
      for (p=0; p<dimll[1]-1; p++){
	if (eps2<(sum+pweights[c][p]))
	  break;
	sum += pweights[c][p];
      }
      //eps2 = (eps2 - sum) / pweights[c][p];
      
      float phi = (c+eps1)/dimll[0]*2.0*M_PI;
      phi -= M_PI_2;
      if (phi<0.0)
	phi += M_PI*2.0;
      
      float theta = (p+eps2)/dimll[1]*M_PI;

      dir.setValue(sin(theta)*cos(phi),
		   cos(theta),
		   sin(theta)*sin(phi));

      col = getRadiance(dir);

      // return pdf
      return(cweights[c] * pweights[c][p] * dimll[0] * dimll[1] /
	      (2.0 * M_PI * M_PI * sint[p]));
}

SbColor EnvironMap::getRadiance(SbVec3f &dir){

  // u,v are [-1,1]
  int i, j;
  float u, v, r;

  // rotate direction
  SbVec3f nDir;
  worldToTextureRot.multDirMatrix(dir, nDir);

  // sometimes this happens because of floating point errors
  if (nDir[Z]>1.0)
    nDir[Z]=1.0;
  else if (nDir[Z]<-1.0)
    nDir[Z]=-1.0;

  // center of image is positive z dir ?
  r = 1.0/M_PI*acos(nDir[Z])/sqrt(nDir[X]*nDir[X]+nDir[Y]*nDir[Y]);
  u = nDir[X]*r;
  v = nDir[Y]*r;

  // translate u,v to [0,1]
  u = (u+1.0)/2.0;
  v = (v+1.0)/2.0;

  
//   // translate to pixel coords
//   i = (int)(u*dim[0]);
//   j = (int)(v*dim[1]);

//   return (data[dim[0]*j+i]);
  
  // calculate texture coordinates
  float i_f, j_f;
  i_f = (dim[0]-1)*u;
  j_f = (dim[1]-1)*v;

  // calculate indices
  //int i, j;
  int i1, j1;
  i  = (int)floor(i_f);
  j  = (int)floor(j_f);
  i  = clamp( i, 0, dim[0]-2);
  j  = clamp( j, 0, dim[1]-2);
  i1 = i+1;
  j1 = j+1;
	 
  // calculate texel ratios
  float sp, tp;
  sp = i_f-i;
  tp = j_f-j;
	 
  // perform hermite smoothing
  //sp = 3.0*sp*sp-2.0*sp*sp*sp;
  //tp = 3.0*tp*tp-2.0*tp*tp*tp;

  SbColor retVal =
      (1.0-sp)*(1.0-tp)*data[dim[0]*j +i ] +
            sp*(1.0-tp)*data[dim[0]*j +i1] +
      (1.0-sp)*tp      *data[dim[0]*j1+i ] +
            sp*tp      *data[dim[0]*j1+i1];

//   if (!finite(retVal[R]) ||
//       !finite(retVal[G]) ||
//       !finite(retVal[B]) )
//     cerr << "badval at (i,j) " << i << "," << j << "!\n";
    
  return (retVal);
}


SbColor EnvironMap::volImpSample(Ray &r, int samples){

  /////////////////////////////////
  // importance sample the lights
  /////////////////////////////////

  SbColor color(0.0, 0.0, 0.0);

//   // stratified sampling
//   int sqrt_samples = (int)sqrt((float)gv->lightSamples*2);
  
//   // use one sample in final gather
//   if (simple)
//    sqrt_samples = 1;


  // FIXME: this sqrt is wasteful. precompute it
  int sqrt_samples = (int)sqrt((float)samples);
  
  for (int sampx=0; sampx<sqrt_samples; sampx++){
    for (int sampy=0; sampy<sqrt_samples; sampy++){

      // jittered
      double eps1 = (sampx + r.ctx->mydrand48()) / sqrt_samples;
      double eps2 = (sampy + r.ctx->mydrand48()) / sqrt_samples;

      // pick column
      double sum = 0., sumNext = 0.;
      int c, lastGood = 0;
      for (c=0; c<dimll[0]-1; c++){
        if (cweights[c]<=0) continue;
	lastGood = c;
        sumNext += cweights[c];
	if (eps1<=sumNext) break;
	sum = sumNext;
      }

      if (cweights[c]<=0.0) c = lastGood;

      eps1 = (eps1 - sum) / cweights[c];

      // pick row
      sum = sumNext = 0.0;
      lastGood = 0;
      int p;
      for (p=0; p<dimll[1]-1; p++){
        if (pweights[c][p]<=0) continue;
	lastGood = p;
        sumNext += pweights[c][p];
	if (eps2<=sumNext) break;
	sum = sumNext;
      }
      if (pweights[c][p]<=0.0) p = lastGood;
      
      eps2 = (eps2 - sum) / pweights[c][p];

      float phi = (c+eps1)/dimll[0]*2.0*M_PI;
      phi -= M_PI_2;
      if (phi<0.0)
	phi += M_PI*2.0;
      
      float theta = (p+eps2)/dimll[1]*M_PI;

      SbVec3f l;
      l.setValue(sin(theta)*cos(phi),
		 cos(theta),
		 sin(theta)*sin(phi));
      
      // trace from point to light point
      // i add a small amount to make it not intersect overlapping surfaces
      // i really hate doing this
      // optimize: fix this vv^^
      // fix: could solve with min_t
      const float t_eps = 0.002;
      Ray lr = Ray(r.x+l*t_eps, l, r.ctx);
      lr.ignore(r.prim);
      if (gv->pane->scenePrim->intersectRay(lr)){
	continue; // hits something (and not background)
      }

      SbVec3f Le = getRadiance(l);
      float pw = cweights[c] * pweights[c][p] * dimll[0] * dimll[1] /
	(2.0 * M_PI * M_PI * sin(theta));
      float value = 1.0 / pw;

      // ray marching absorbtion
      if (gv->stepSize>0){
	float tau = 0.0;
	float sdt = gv->stepSize;
	float sigma_t = 0.0;
	SbVec3f lx = r.x+gv->pane->getMaxDist()*l;
	tau = gv->pane->dataSet->bresenSum(r.x, lx);
	tau = gv->meters(tau);
	value *= exp(-tau);
      }
	  
      // compute color
      const float pp = (1./4.)*M_1_PI;
      color += Le * value;// * pp;

    } // end stratified sampling y
  } // end stratified sampling x

  color /= sqrt_samples*sqrt_samples;
      
  // debugging
  //writeData3();
  
  return(color);
}

void EnvironMap::setRotateSky(const SbRotation &rot){
  textureToWorldRot.setRotate(rot);
  worldToTextureRot = textureToWorldRot.inverse();
}

void EnvironMap::writeDataExr(const string &file){
  FileUtils::writeEXR(dim, file, data);
  cerr << "Wrote environment map " << file << endl;
}

void EnvironMap::writeDataHdr(const string &file){
  FileUtils::writeHDR(dim, file, data);
  cerr << "Wrote environment map " << file << endl;
}

SbColor EnvironMap::getEmissiveFlux(float receiverArea){
  // L = Phi/(Area*cosTheta*Omega) -> Phi = L*Area*cosTheta*Omega
  return Lavg * receiverArea * 4.0*M_PI; // cosTheta = 1
}
