// Header for Misc.cxx

#ifndef MISCFUNCS
#define MISCFUNCS

#include <float.h>
#include <Inventor/SbLinear.h>

class SoNode;


class Vec3d {
public:

  double vec[3];          // position, also color (r,g,b)
    
  Vec3d(double x_=0, double y_=0, double z_=0){
    vec[0] = x_;
    vec[1] = y_;
    vec[2] = z_;
  }
  
  Vec3d(const SbVec3f &v) {
    vec[0] = v[0];
    vec[1] = v[1];
    vec[2] = v[2];
  }
  
  operator SbVec3f() {
    return SbVec3f(vec[0], vec[1], vec[2]);
  }
  
  Vec3d operator+(const Vec3d &b) const {
    return Vec3d(vec[0]+b.vec[0],
                 vec[1]+b.vec[1],
                 vec[2]+b.vec[2]);
  }
  
  Vec3d operator-(const Vec3d &b) const {
    return Vec3d(vec[0]-b.vec[0],
                 vec[1]-b.vec[1],
                 vec[2]-b.vec[2]);
  }
  
  Vec3d operator*(double b) const {
    return Vec3d(vec[0]*b,
                 vec[1]*b,
                 vec[2]*b);
  }
  
  Vec3d& operator*=(double b) {
    vec[0] *= b;
    vec[1] *= b;
    vec[2] *= b;
    return *this;
  }
  
  Vec3d scale(const Vec3d &b) const {
    return Vec3d(vec[0]*b.vec[0],
                 vec[1]*b.vec[1],
                 vec[2]*b.vec[2]);
  }
  
  Vec3d norm(){
    double len2 = vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2];
    double len;
    Vec3d n(0,0,0);
    if (!len2) return n;
    len = sqrt(len2);
    n = *this * (1/len);
    return n;
  }

  double normalize(){
    double len2 = vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2];
    double len;
    if (len2){
      len = sqrt(len2);
      *this = *this * (1/len);
    }
    return len;
  }
  
  double length(){
    return sqrt(vec[0]*vec[0] +
                vec[1]*vec[1] +
                vec[2]*vec[2]);
  }
  
  double dot(const Vec3d &b) const {
    return
      vec[0]*b.vec[0] +
      vec[1]*b.vec[1] +
      vec[2]*b.vec[2];
  }
  
  // cross prod  
  Vec3d operator%(const Vec3d &b) const {
    return Vec3d(vec[1]*b.vec[2] - vec[2]*b.vec[1],
                 vec[2]*b.vec[0] - vec[0]*b.vec[2],
                 vec[0]*b.vec[1] - vec[1]*b.vec[0]);
  }
  
  Vec3d cross(const Vec3d &b) const { return *this % b; }

  double       & operator [](int i) 	  { return vec[i]; }
  const double & operator [](int i) const { return vec[i]; }
};


int computeNormals(SoNode *scene);


//'WWW''WW'''W'WWWW''''''''''WWW''''''''''''''''''
//WW  W WWW  W WW  W          WW                  
//WW  W W WW W WWWW      WWW  WW   WWW  WWWW  WWWW
//WW  W W  WWW WW  W    WW W  WW  W WW WWW   WWW  
//WW  W W   WW WW  W    WW    WW  W WW   WWW   WWW
// WWW  W    W WWWW      WWW WWWW  WWW WWWW  WWWW 
//                                                
//''''''''''''''''''''''''''''''''''''''''''''''''

// ortho-normal basis class
class ONB {

private:
  
  SbVec3f u, v, w; // w "points up"

public:

  ONB(){}
  ONB(SbVec3f &c) {
    constructFromW(c);
  }
  
  void init(){
    u.setValue(1,0,0);
    v.setValue(0,1,0);
    w.setValue(0,0,1);
  }

  // from two triangle edges
  void constructFromUV(SbVec3f &a, SbVec3f &b){
    u = a;
    u.normalize();

    w = a.cross(b);
    w.normalize();
    
    v = w.cross(u);
    v.normalize();
  }

  // from up direction
  void constructFromW(SbVec3f &c){
    w = c;
    w.normalize();

    u = SbVec3f(w[1], w[2], w[0]+0.5); // Hopefully this !~ w
    u -= u.dot(w)*w;
    u.normalize();

    v = w.cross(u);
    v.normalize();
  }

  // from U and W directions
  void constructFromUW(SbVec3f &a, SbVec3f &c){
    w = c;
    w.normalize();

    u = a;
    u -= u.dot(w)*w;
    u.normalize();

    v = w.cross(u);
    v.normalize();
  }
  
  SbVec3f toONBSpace(const SbVec3f &a) const {
    SbVec3f x;
    x[0] = u.dot(a);
    x[1] = v.dot(a);
    x[2] = w.dot(a);
    return x;
  }

  Vec3d toONBSpace(const Vec3d &a) const {
    Vec3d x;
    x[0] = a.dot(u); //u.dot(a);
    x[1] = a.dot(v); //v.dot(a);
    x[2] = a.dot(w); //w.dot(a);
    return x;
  }
  
  SbVec3f toWorldSpace(const SbVec3f &a) const {
    return SbVec3f(a[0]*u + a[1]*v + a[2]*w);
  }
  
  Vec3d toWorldSpace(const Vec3d &a) const {
    return Vec3d(Vec3d(u)*a[0] + Vec3d(v)*a[1] + Vec3d(w)*a[2]);
  }

  const SbVec3f& getU() const { return u; }
  const SbVec3f& getV() const { return v; }
  const SbVec3f& getW() const { return w; }

};

class BBox {

public:
  SbVec3f min, max;

  BBox(){
    min = SbVec3f(FLT_MAX, FLT_MAX, FLT_MAX);
    max = SbVec3f(-FLT_MAX, -FLT_MAX, -FLT_MAX);
  }
  
  BBox(SbVec3f *p, int n){
    BBox::BBox();
    grow(p, n);
  }

  SbVec3f size(){
    return max-min;
  }
  
  void grow(const SbVec3f *vert, int n){
    for (int v=0; v<n; v++){
      for (int d=0; d<3; d++){
        if (vert[v][d] < min[d])
          min[d] = vert[v][d];
        if (vert[v][d] > max[d])
          max[d] = vert[v][d];
      }
    }
  }

  void grow(const BBox &x) {
    grow(&x.min, 1);
    grow(&x.max, 1);
  }
  
  BBox formUnion(const BBox &x) const {
    BBox u = *this;
    u.grow(x);
    return u;
  }

  static SbVec3f vecMin(const SbVec3f &a, const SbVec3f &b) {
    SbVec3f min;
    for (int i=0; i<3; i++)
      min[i] = a[i]<b[i] ? a[i] : b[i];
    return min;
  }
  
  static SbVec3f vecMax(const SbVec3f &a, const SbVec3f &b) {
    SbVec3f max;
    for (int i=0; i<3; i++)
      max[i] = a[i]>b[i] ? a[i] : b[i];
    return max;
  }
  
};

#endif /* MISCFUNCS */
