#include<stack>
#include <fenv.h>
#include "Octree.h"
#include "debug.h"
#include "Ray.h"
#include "Misc.h"
#include "Main.h"

const int Octree::delta[8][3] = {
  0, 0, 0,
  0, 0, 1,
  0, 1, 0,
  0, 1, 1,
  
  1, 0, 0,
  1, 0, 1,
  1, 1, 0,
  1, 1, 1
};
  

Octree::Octree(){
  min[X] = min[Y] = min[Z] = 1e99;
  max[X] = max[Y] = max[Z] = -1e99;
  shiny_min.setValue(1e99f, 1e99f, 1e99f);
  shiny_max.setValue(-1e99f, -1e99f, -1e99f);
  root = new node;
  for (int i=0; i<8; i++)
    root->child[i] = NULL;
  root->primList = &primList;
  root->mid[0] = root->mid[1] = root->mid[2] = 0.0;
  numPrims = 0;
  maxSideLength = 0.0;
  octreeDebug = 0;
  maxDepth = 8;
  maxItems = 10;
  rayId = 0;
  pthread_mutex_init(&lock, NULL);
  rays = 0;
  compactNodes = NULL;
  compactNodesSize = 0;
}

// recursive deletes nodes/children of nodes (but not triangles)
void Octree::deleteNode(node *n){
  if (n == NULL)
    return;

  // delete children
  if (n->primList)
    if (n != root)  // root's is Octree member
      delete n->primList;
  else
    for (int i=0; i<8; i++)
      deleteNode(n->child[i]);

  // delete self
  delete n;
}

Octree::~Octree(){
  struct timeval start, end;
  cerr << "Deleting triangle tree..." << flush;
  gettimeofday(&start, NULL); // start delay measure

  // delete nodes
  cerr << "(octree nodes...)" << flush;
  if (!compactNodes)
    deleteNode(root);
  else {
    for (int i=1; i<compactNodesSize; i++)
      delete compactNodes[i].primList;
    delete [] compactNodes;
    compactNodes = 0;
  }

  // destroy mutex
  pthread_mutex_destroy( &lock );

  gettimeofday(&end, NULL);   // end delay measure
  cerr << " done. (" << delay(start, end) << "ms)" << endl;
}


// Resets tree, deleting all nodes
// This is called by rescanScene() to regenerate an octree
// because of changed geometry.
void Octree::reset(){

  // clear tree visualization
  showCube(SbVec3f(0.0, 0.0, 0.0), SbVec3f(0.0, 0.0, 0.0),
	   SbVec3f(0.0, 0.0, 0.0), 1);

  // clear tree
  primList.clear();
  deleteNode(root);
  root = NULL;

  // reset variables
  root = new node;
  for (int i=0; i<8; i++){
    root->child[i] = NULL;
  }
  root->primList = &primList;
  root->mid[0] = root->mid[1] = root->mid[2] = 0.0;
  min[X] = min[Y] = min[Z] = 1e99;
  max[X] = max[Y] = max[Z] = -1e99;
  shiny_min.setValue(1e99f, 1e99f, 1e99f);
  shiny_max.setValue(-1e99f, -1e99f, -1e99f);
  maxSideLength = 0.0;
  numPrims = 0;
}
  
// set triangle tree parameters
void Octree::setParams(int maxD, int maxI, int debug){
  maxDepth = maxD;
  maxItems = maxI;
  octreeDebug = debug;
}

// creates the octree structure
void Octree::refine(int simpleDiv){
    
  // adjust first node boundaries
  // (make global bounding box)
  for (int ti=0; ti<primList.size(); ti++){

    const Primitive *tri = primList.prim(ti);
    min = BBox::vecMin(min, tri->getMin());
    max = BBox::vecMax(max, tri->getMax());
    shiny_min = BBox::vecMin(shiny_min, tri->getMinShiny());
    shiny_max = BBox::vecMax(shiny_max, tri->getMaxShiny());
  }

  // setup boundaries of first node
  for (int i=0; i<3; i++){
    min[i] -= EPSILON_OS;
    max[i] += EPSILON_OS;
    shiny_min[i] -= EPSILON_OS;
    shiny_max[i] += EPSILON_OS;
    if (max[i]<min[i]){
      min[i] = -1;
      max[i] = 1;
    }
    if (shiny_max[i]<shiny_min[i]){
      shiny_min[i] = -1;
      shiny_max[i] = 1;
    }
    root->mid[i] = (min[i]+max[i])/2.0;
    if (max[i]-min[i] > maxSideLength)
      maxSideLength = max[i]-min[i];
  }

  // find bounding sphere
  center = SbVec3f(min[X]+max[X], min[Y]+max[Y], min[Z]+max[Z])/2.0;
  SbVec3f maxCorner(max[X], max[Y], max[Z]);
  radius = (maxCorner-center).length();
  //showPoint(center, radius, SbColor(0.0, 0.0, 1.0));

  // shiny bounding sphere
  shiny_center = (shiny_min+shiny_max)/2.0;
  shiny_radius = (shiny_max-shiny_center).length();
  //showPoint(shiny_center, shiny_radius, SbColor(1.0, 0.0, 0.0));
  
#ifdef SHOWTRAV  
  // debugging
  root->min[0] = min[0];
  root->min[1] = min[1];
  root->min[2] = min[2];
  root->max[0] = max[0];
  root->max[1] = max[1];
  root->max[2] = max[2];
  root->depth = 0;
#endif
  
  if (min[X] == max[X] ||
      min[Y] == max[Y] ||
      min[Z] == max[Z]){
    cerr << "You forgot to set the size of the tree before refine().\n";
    exit(1);
  }
    
  refineNode(root, 0, simpleDiv, min, max);

  if (octreeDebug){
    showCube(SbVec3f(min[X], min[Y], min[Z]),
	     SbVec3f(max[X], max[Y], max[Z]),
	     SbColor(0.0, 1.0, 1.0));
//    showCube(SbVec3f(-0.5, -0.5, -0.5),
//	     SbVec3f(0.5, 0.5, 0.5),
//	     SbColor(0.0, 0.0, 1.0));
  }

  reorderNodes();
}


const static int mod3[5] = { 0, 1, 2, 0, 1 };

double Octree::findDividingPlane(node *n, int a,
                                 const SbVec3f &vmin, const SbVec3f &vmax,
                                 const vector<BBox> &bboxs){
  
  int maxErr = MAX2((int)(0.05*n->primList->size()), 1);
  int splits = 0;
  double x0 = vmin[a], x1 = vmax[a];
  int c0, c1, cb;
  int diff = maxErr+10; // guarantees >=1 iteration, which inits this correctly
  double xm = (vmin[a]+vmax[a])*0.5; // middle point
  double xn;                         // median point
  
  // binary search for median
  while (diff>maxErr && splits++<10){
    
    xn = (x0+x1)*0.5;

    // count triangles less than and greater than xn
    c0 = c1 = 0;
    for (int ti=0; ti<n->primList->size(); ti++){

      const SbVec3f &myMin = bboxs[ti].min;
      const SbVec3f &myMax = bboxs[ti].max;
      
      int r = myMax[a] >  xn ? 1 : 0;
      int l = myMin[a] <= xn ? 1 : 0;
      if (l&&r)	continue;
      c0 += l;
      c1 += r;
    }

    if (c0>=c1)
      x1 = xn;
    else
      x0 = xn;
    diff = abs(c1-c0);
  }


  // minimize E from equation 8 in Octree-R paper
  int divs = 10;

  double va = vmax[a]-vmin[a];
  double vb = vmax[mod3[a+1]]-vmin[mod3[a+1]];
  double vc = vmax[mod3[a+2]]-vmin[mod3[a+2]];

  //cerr << "va,vb,vc = " << va << ", " << vb << ", " << vc << "\n";

  double min_e = 1e99;
  double min_x = xm;

  // loop over points between xm and xn
  for (int k=0; k<=divs; k++){
    double x = (xn-xm)/divs * k + xm;
    double t = x-vmin[a];

    // count triangles less than and greater than x
    c0 = c1 = cb = 0;
    for (int ti=0; ti<n->primList->size(); ti++){

      const SbVec3f &myMin = bboxs[ti].min;
      const SbVec3f &myMax = bboxs[ti].max;
      
      int r = myMax[a] >  x ? 1 : 0;
      int l = myMin[a] <= x ? 1 : 0;

      if (l&&r)	  cb++;
      else if (l) c0++;
      else        c1++;
    }

    double d = va*(vb+vc)+vb*vc;
    double e =
      (     t*(vb+vc)+vb*vc)/d      * c0 +
      ((va-t)*(vb+vc)+vb*vc)/d      * c1 +
      (    va*(vb+vc)+2.0*vb*vc)/d  * cb;

    if (e<min_e){
      min_e = e;
      min_x = x;
    }
    
  }
  
  return(min_x);
}


void Octree::refineNode(node *n, int depth, int simpleDiv,
                        const SbVec3f &vmin, const SbVec3f &vmax){

  SbMatrix scale;
  float nmin[3], nmax[3];

  // precompute (once...) bounding boxes
  vector<BBox> bboxs;
  if (!simpleDiv){
    int s = n->primList->size();
    bboxs.resize(s);
    for (int i=0; i<s; i++){
      bboxs[i].min = n->primList->prim(i)->getMin();
      bboxs[i].max = n->primList->prim(i)->getMax();
    }
  }

  // subdivide node
  if (depth<maxDepth && n->primList && n->primList->size()>maxItems){
    int i;

    // debugging info
    if (0 && octreeDebug){
      for (i=0; i<depth; i++)
	cerr << "  ";
      cerr << "refineNode(" << n << ") "
	   << "(" << vmin[X] << ", " << vmax[X] << ") "
	   << "(" << vmin[Y] << ", " << vmax[Y] << ") "
	   << "(" << vmin[Z] << ", " << vmax[Z] << ")"
	   << " depth=" << depth
	   << " items=" << n->primList->size() << endl;;
    }

    // set new size
    if (simpleDiv){        // midpoint 
      n->mid[X] = 0.5*(vmax[X]-vmin[X])+vmin[X]; 
      n->mid[Y] = 0.5*(vmax[Y]-vmin[Y])+vmin[Y];
      n->mid[Z] = 0.5*(vmax[Z]-vmin[Z])+vmin[Z];
    }
    else {                 // optimized point for subdivision
      n->mid[X] = findDividingPlane(n, X, vmin, vmax, bboxs);
      n->mid[Y] = findDividingPlane(n, Y, vmin, vmax, bboxs);
      n->mid[Z] = findDividingPlane(n, Z, vmin, vmax, bboxs);

      // free memory
      vector<BBox>().swap(bboxs);
    }

    // for each child
    int childCreated = 0;
    for (i=0; i<8; i++){

      childCreated = 0;

      nmin[X] = delta[i][X] ? n->mid[X] : vmin[X]; 
      nmin[Y] = delta[i][Y] ? n->mid[Y] : vmin[Y]; 
      nmin[Z] = delta[i][Z] ? n->mid[Z] : vmin[Z];
      nmax[X] = delta[i][X] ? vmax[X] : n->mid[X];
      nmax[Y] = delta[i][Y] ? vmax[Y] : n->mid[Y];
      nmax[Z] = delta[i][Z] ? vmax[Z] : n->mid[Z];

      // debugging info: show octree/child cell
      SbColor color;
      color.setHSVValue(clamp((double)n->primList->size(),
			      (double)0.0,(double)36.0)/36.0,
			1.0, 0.5);

      // add appropriate children to new child node
      //   run triangle cube intersection for each triangle
      double boxcenter[3], boxhalfsize[3];
      boxcenter[0] = (nmax[X]+nmin[X])/2.0;
      boxcenter[1] = (nmax[Y]+nmin[Y])/2.0;
      boxcenter[2] = (nmax[Z]+nmin[Z])/2.0;
      boxhalfsize[0] = nmax[X]-boxcenter[0];
      boxhalfsize[1] = nmax[Y]-boxcenter[1];
      boxhalfsize[2] = nmax[Z]-boxcenter[2];

      // test each triangle
      int children = 0;
      int tri_num = 0;
      for (int ti=0; ti<n->primList->size(); ti++){
	Primitive *tri = n->primList->prim(ti);

	tri_num++;

	if (tri->intersectCubeCenterHalf(boxcenter, boxhalfsize)){
	
	  if (!childCreated){
	    // create child
	    n->child[i] = new node;
	    for (int j=0; j<8; j++)
	      n->child[i]->child[j] = NULL;
	    n->child[i]->primList = new PrimList();
	    n->child[i]->mid[0] = boxcenter[0];
	    n->child[i]->mid[1] = boxcenter[1];
	    n->child[i]->mid[2] = boxcenter[2];
	    childCreated++;

#ifdef SHOWTRAV	    
	    // debugging
	    n->child[i]->min = SbVec3f(nmin[X], nmin[Y], nmin[Z]);
	    n->child[i]->max = SbVec3f(nmax[X], nmax[Y], nmax[Z]);
	    n->depth=depth+1;
#endif	    

	    if (octreeDebug)
	      showCube(SbVec3f(nmin[X], nmin[Y], nmin[Z]),
		       SbVec3f(nmax[X], nmax[Y], nmax[Z]),
		       color);
	  }
	    
	  n->child[i]->primList->add(tri);
	  children++;
	}
      }

      // refine new child node node
      if (childCreated)
	refineNode(n->child[i], depth+1, simpleDiv, nmin, nmax);
      else
	n->primList->resizeToFit(); // save memory
	
    } // end for(1..8)

      // remove Primitives from primList, set terminal to 0
      n->primList->clear();
      if (n != root){            // root node's is member var
        delete n->primList;
      }
      n->primList = NULL;
      
  } // end if (depth<max_depth...)
    
  else{ // if (n->primList.size()){

    // debugging info: print cell info
    if (0 && octreeDebug){
      if (1 || n->primList->size()){ // all or just empty cells
	for (int x=0; x<depth; x++)
	  cerr << "  ";
	cerr << n << " ";
	cerr << n->primList->size() << " triangles depth=" << depth << endl;
      }
    }
      
  }

} // end of refineNode();


  /*

  This Octree Traversal code is taken from
  "An Efficient Parametric Algorithm for Octree Traversal"
  by J. Revelles, C. Urena, M. Lastra

  */

  // returns i-th int, where i'th float is minimum
inline int Octree::new_node(double &t1, int i1, double &t2, int i2,
				  double &t3, int i3) const{
  return( (t1<t2&&t1<t3) ? i1 : (t2<t3 ? i2 : i3) );
}

// returns first child node crossed
int Octree::first_node(double &tx0, double &ty0, double &tz0,
			     double &txm, double &tym, double &tzm) const{
    
  unsigned char b = 0;
  if (tx0>ty0 && tx0>tz0){   // tx0 max
    if (tym<tx0)
      b |= 2;
    if (tzm<tx0)
      b |= 1;
  }
  else if (ty0>tz0){         // ty0 max
    if (txm<ty0)
      b |= 4;
    if (tzm<ty0)
      b |= 1;
  }
  else {                     // tz0 max
    if (txm<tz0)
      b |= 4;
    if (tym<tz0)
      b |= 2;
  }

  return(b);
}

bool Octree::intersectRay(Ray &r, double t0, double t1) {

  // count rays
  r.ctx->rays++;

  // set no hit by default
  r.t = -1;
  r.prim = NULL;
  
  // return value pointers (private class members)
  treeParam tp;
  tp.r_p = &r;
  tp.t0 = MAX2(t0,EPSILON_OI); // is this MAX(,EPS) op necessary?
  tp.t1 = t1;
  disableFPExceptions();
  tp.invDir = SbVec3f(1/r.dir[X], 1/r.dir[Y], 1/r.dir[Z]);

  // function variables
  SbVec3f pos = r.pos;
  SbVec3f invDir = tp.invDir;
  double refl[3];
  double tx0, tx1, ty0, ty1, tz0, tz1;

  // calculate midplanes
  refl[0] = (max[0]+min[0]);
  refl[1] = (max[1]+min[1]);
  refl[2] = (max[2]+min[2]);
    
  tp.a = 0;
  if (invDir[X]<0.0){
    pos[X] = refl[X]-pos[X];
    invDir[X] = -invDir[X];
    tp.a |= 4;
  }
  if (invDir[Y]<0.0){
    pos[Y] = refl[Y]-pos[Y];
    invDir[Y] = -invDir[Y];
    tp.a |= 2;
  }
  if (invDir[Z]<0.0){
    pos[Z] = refl[Z]-pos[Z];
    invDir[Z] = -invDir[Z];
    tp.a |= 1;
  }

  tx0 = (min[X]-pos[X])*invDir[X];
  tx1 = (max[X]-pos[X])*invDir[X];
  ty0 = (min[Y]-pos[Y])*invDir[Y];
  ty1 = (max[Y]-pos[Y])*invDir[Y];
  tz0 = (min[Z]-pos[Z])*invDir[Z];
  tz1 = (max[Z]-pos[Z])*invDir[Z];
  enableFPExceptions();
  
  if (MAX3(tx0,ty0,tz0) >= MIN3(tx1,ty1,tz1))
    return false;
  
  return proc_subtree(tx0,ty0,tz0, tx1,ty1,tz1, root, tp);
}

bool Octree::proc_subtree(double &tx0, double &ty0, double &tz0,
                          double &tx1, double &ty1, double &tz1,
                          node *n, treeParam &tp) const{

  double txm, tym, tzm;
  unsigned int currNode;

  // don't examine volumes behind the position
  if (n == NULL || tx1<0.0 || ty1<0.0 || tz1<0.0)
    return false;

#ifdef SHOWTRAV  
  // debugging
  if (DEBUG_STATE){
    cerr << "Size: " << n->max-n->min << "\n";
    SbColor color;
    color.setHSVValue(n->depth/8.0, 1.0, 0.5);
    showCube(n->min, n->max, color);
  }
#endif
  
  // proc_terminal(n);
  if (n->primList != NULL){
    double t1 = MIN3(tx1,ty1,tz1)+EPSILON_OI;
    t1 = MIN2(t1, tp.t1);
    return n->primList->intersectRay(*tp.r_p, tp.t0, t1);
  }

  // this code is for axis-aligned direction handling (special case)
  // see paper Sec.3.3 in An Efficient Parametric Algorithm for Octree Traversal
  if (isinf(tx1) == 1){
    if (tp.r_p->pos[X] < n->mid[X])
      txm = tx1;
    else
      txm = -tx1;
  }
  else
    txm = (n->mid[X]-tp.r_p->pos[X])*tp.invDir[X];

  if (isinf(ty1) == 1){
    if (tp.r_p->pos[Y] < n->mid[Y])
      tym = ty1;
    else
      tym = -ty1;
  }
  else
    tym = (n->mid[Y]-tp.r_p->pos[Y])*tp.invDir[Y];
  
  if (isinf(tz1) == 1){
    if (tp.r_p->pos[Z] < n->mid[Z])
      tzm = tz1;
    else
      tzm = -tz1;
  }
  else
    tzm = (n->mid[Z]-tp.r_p->pos[Z])*tp.invDir[Z];
  
  currNode = 0;
  if (tx0>ty0 && tx0>tz0){   // tx0 max
    if (tym<tx0)
      currNode |= 2;
    if (tzm<tx0)
      currNode |= 1;
  }
  else if (ty0>tz0){         // ty0 max
    if (txm<ty0)
      currNode |= 4;
    if (tzm<ty0)
      currNode |= 1;
  }
  else {                     // tz0 max
    if (txm<tz0)
      currNode |= 4;
    if (tym<tz0)
      currNode |= 2;
  }
  
  do {
    switch (currNode){
    case 0:
      if ( proc_subtree(tx0, ty0, tz0, txm, tym, tzm, n->child[tp.a], tp) )
	return true;
      currNode = new_node(txm, 4, tym, 2, tzm, 1);
      break;
	
    case 1:
      if ( proc_subtree(tx0, ty0, tzm, txm, tym, tz1, n->child[1^tp.a], tp) )
	return true;
      currNode = new_node(txm, 5, tym, 3, tz1, 8);
      break;

    case 2:
      if ( proc_subtree(tx0, tym, tz0, txm, ty1, tzm, n->child[2^tp.a], tp) )
	return true;
      currNode = new_node(txm, 6, ty1, 8, tzm, 3);
      break;

    case 3:
      if ( proc_subtree(tx0, tym, tzm, txm, ty1, tz1, n->child[3^tp.a], tp) )
	return true;
      currNode = new_node(txm, 7, ty1, 8, tz1, 8);
      break;

    case 4:
      if ( proc_subtree(txm, ty0, tz0, tx1, tym, tzm, n->child[4^tp.a], tp) )
	return true;
      currNode = new_node(tx1, 8, tym, 6, tzm, 5);
      break;

    case 5:
      if ( proc_subtree(txm, ty0, tzm, tx1, tym, tz1, n->child[5^tp.a], tp) )
	return true;
      currNode = new_node(tx1, 8, tym, 7, tz1, 8);
      break;

    case 6:
      if ( proc_subtree(txm, tym, tz0, tx1, ty1, tzm, n->child[6^tp.a], tp) )
	return true;
      currNode = new_node(tx1, 8, ty1, 8, tzm, 7);
      break;

    case 7:
      if ( proc_subtree(txm, tym, tzm, tx1, ty1, tz1, n->child[7^tp.a], tp) )
	return true;
      currNode = 8;
      break;
	
    default:
      //cerr << "Invalid case " << currNode << " in switch(currNode)!" << endl;
      break;
    } // end switch
      
  } while (currNode<8); // end do {} while

  return false;
} // end proc_subtree()

// single sided intersection routine
//   calls regular intersection routine
bool Octree::ss_intersectRay(Ray &r, double t0, double t1){


  // this sucks and should be modified. doesn't even look like it works..
  bool retVal = intersectRay(r, t0, t1);
    
  if (retVal && r.prim->getNormal(r.bc).dot(r.dir)>0.0){
    Ray new_r = Ray(r.pos+(r.t+0.0001)*r.dir, r.dir, r.ctx);
    retVal = ss_intersectRay(new_r, t0, t1-r.t);
    r.t += new_r.t;
  }

  return retVal;
}

void Octree::printStats(){

  int ncubes = 1, ntris = 0;
  printStats(root, &ncubes, &ntris);
  cerr << "Octree statistics: "
       << ncubes << " cubes, "
       << ntris << " triangles, "
       << numPrims << " initially ("
       << ntris/float(numPrims?numPrims:1)*100.0 << "%).\n";
  SbVec3f eps(EPSILON_OS, EPSILON_OS, EPSILON_OS);
  cerr << "  min:(" << min+eps << ")  max=(" << max-eps << "\n";
  return;
}

void Octree::printStats(node *ptr, int *numCubes, int *numTris){

  if (ptr == NULL) return;

  // if leaf node, count tets
  if (ptr->primList != NULL){
    *numTris += ptr->primList->size();
  }

  // otherwise, recurse over children
  else {
    for (int i=0; i<8; i++){
      if (ptr->child[i] != NULL){
	(*numCubes)++;
	printStats(ptr->child[i], numCubes, numTris);
      }
    }
  }

  return;
}


void Octree::reorderNodes(){
  
  // count nodes
  node *p;
  int nodeCnt = 0;
  stack<node *> nodes;
  nodes.push(root);

  while (nodes.empty() == 0){
    
    p = nodes.top();
    nodes.pop();
    nodeCnt++;
    
    for (int i=0; i<8; i++)
      if (p->child[i])
	nodes.push(p->child[i]);
  }

  cerr << "Counted " << nodeCnt << " nodes.\n";

  // allocate space for nodes
  compactNodes = new node[nodeCnt];
  compactNodesSize = nodeCnt;
  int nodeIdx = 0;
  node *q;
  
  compactNodes[nodeIdx] = *root;
  delete root;
  root = &compactNodes[nodeIdx++];
  nodes.push(root);

  while (nodes.empty() == 0){
    
    p = nodes.top(); // p already in compactNodes
    nodes.pop();

    // move p's children
    for (int i=0; i<8; i++)
      if (p->child[i]){

	// copy node
	compactNodes[nodeIdx] = *(p->child[i]);
	delete p->child[i];
	p->child[i] = &compactNodes[nodeIdx++];

	// add node to processing list
	nodes.push(p->child[i]);
      }
  }

  cerr << "Moved " << nodeIdx << " nodes.\n";

  return;
}
