#ifndef PRIMITIVE_H
#define PRIMITIVE_H

#include<float.h>
#include<vector>
#include "Ray.h"
#include "Misc.h"
#include "Material.h"

using namespace std;

class SbVec3f;

typedef struct PrimDerivs {
  SbVec3f dPdU;
  SbVec3f dPdV;
  SbVec3f dNdU;
  SbVec3f dNdV;
  PrimDerivs() {}
  PrimDerivs(const SbVec3f &a, const SbVec3f &b,
             const SbVec3f &c, const SbVec3f &d) :
    dPdU(a), dPdV(b), dNdU(c), dNdV(d) {}
} PrimDerivs;

class Primitive {

public:

  static const SbColor Black;
  static const SbColor White;
  static const SbVec3f XDir, YDir, ZDir;
  static const PrimDerivs DefaultDerivs;
  static Material defaultMat;

  typedef unsigned short PrimType;
  enum {
    PRIMTYPE_NONE          = 0x00000000,  // 0
    PRIMTYPE_TRIANGLE      = 0x00000001,  // 1
    PRIMTYPE_EMISTRIANGLE  = 0x00000002,  // 2 (1|2 = 3)
    PRIMTYPE_TRIANGLELIST  = 0x00000004,  // 4
    PRIMTYPE_PRIMITIVELIST = 0x00000008,  // 8
    PRIMTYPE_INSTANCE      = 0x00000010,  // 16
    PRIMTYPE_OCTREE        = 0x00000020,  // 32
    PRIMTYPE_SPHERE        = 0x00000040,  // 64
    PRIMTYPE_JULIA         = 0x00000080,  // 128
    PRIMTYPE_PLANE         = 0x00000100,  // 256
  }; // possible values for PrimType

  virtual ~Primitive(){ }
  virtual PrimType getType() const = 0;
  virtual bool isOfType(PrimType p) const { return getType() & p; }
  virtual bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8) = 0;
  virtual bool intersectRaySimple(Ray &r, double t0=1e-5, double t1=1e8) {
    return intersectRay(r, t0, t1);
  }
//  virtual bool intersectCube(const SbVec3f &min, const SbVec3f &max) const;
  virtual bool intersectCube(const double min[3], const double max[3]) const;
  virtual bool intersectCubeCenterHalf(const double center[3],
                                       const double half[3]) const {
    double min[3], max[3];
    min[0] = center[0] - half[0];
    min[1] = center[1] - half[1];
    min[2] = center[2] - half[2];
    max[0] = center[0] + half[0];
    max[1] = center[1] + half[1];
    max[2] = center[2] + half[2];
    return intersectCube(min, max);
  }
  
  virtual SbVec3f  getMin()      const = 0;
  virtual SbVec3f  getMinShiny() const { return getMin(); }
  virtual SbVec3f  getMax()      const = 0;
  virtual SbVec3f  getMaxShiny() const { return getMax(); }

  bool     isShiny()                 const { return getMat()->isShiny();    }
  bool     isGlossy()                const { return getMat()->isGlossy();   }
  bool     isDiffuse()               const { return getMat()->isDiffuse();  }
  bool     isEmissive()              const { return getMat()->emissive;     }
  float    getShininess()            const { return getMat()->getShininess();   }
  float    getGlossiness()           const { return getMat()->getGlossiness();  }
  float    getDiffuseness()          const { return getMat()->getDiffuseness(); }
  float    getTransparency()         const { return getMat()->transparency; }
  const SbColor& getSpecular()       const { return getMat()->specular;     }
  float    getEmisAvg()              const { return getMat()->emisAvg;  }
  SbColor  getEmission()             const { return getMat()->emission; }
  float    getIndexOfRefractionTop() const { return getMat()->ior1;  }
  float    getIndexOfRefractionBot() const { return getMat()->ior2;  }
  const SbColor& getAttenuation()    const { return getMat()->atten; }
  Material::ReflModel getModel()     const { return getMat()->getModel(); }
  float    getRough1()               const { return getMat()->getRough1(); }
  float    getRough2()               const { return getMat()->getRough2(); }
  float    getIsotropy1()            const { return getMat()->getIsotropy1(); }

  virtual Material* getMat() { return &defaultMat; }
  virtual const Material* getMat() const { return &defaultMat; }
  virtual SbVec3f getNormal(      const SbVec3f &bc) const{ return YDir; }
  virtual SbVec3f getIntNormal(   const SbVec3f &bc) const{ return getNormal(bc); }
  virtual SbColor getIntDiffColor(const SbVec3f &bc) const{ return getMat()->diffuse;}
  virtual SbColor getIntEmission( const SbVec3f &bc) const{ return getMat()->emission;}
  virtual bool    getDerivs(      const SbVec3f &bc, PrimDerivs &pDer)const{return 0;}

};

class PrimList : public Primitive {

private:
  vector<Primitive *> prims;

public:
  PrimList(){ }
  ~PrimList(){ }

  int size() const { return prims.size(); }
  const Primitive* prim(int i) const { return &*prims[i]; }
  Primitive* prim(int i) { return &*prims[i]; }
  void add(Primitive *p) { prims.push_back(p); }
  void reserve(int n) { prims.reserve(n); }
  void clear() { prims.clear(); }
  void resizeToFit(){ vector<Primitive *>(prims).swap(prims); }

  virtual PrimType getType() const { return PRIMTYPE_PRIMITIVELIST; }

  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8){
    double minT = -1;
    Primitive *minPrim = NULL;
    SbVec3f minBc(1,0,0);
    bool hit = false;

    // loop over triangles
    for (int i=0; i<prims.size(); i++){

      // skip triangle
      // do this test in child's
      //if (prims[i] == r.ignorePrim) continue;

      if (prims[i]->intersectRay(r, t0, t1)){
        hit = true;
        minT = t1 = r.t;
        minPrim = r.prim;
        minBc = r.bc;
      }
    }

    r.t = minT;
    r.prim = minPrim;
    r.bc = minBc;
    return hit;
  }
  
  virtual SbVec3f  getMin() const {
    SbVec3f myMin(FLT_MAX, FLT_MAX, FLT_MAX);
    for (int i=0; i<prims.size(); i++)
      myMin = BBox::vecMin(myMin, prims[i]->getMin());
    return myMin;
  }
  virtual SbVec3f  getMinShiny() const {
    SbVec3f myMin(FLT_MAX, FLT_MAX, FLT_MAX);
    for (int i=0; i<prims.size(); i++)
      myMin = BBox::vecMin(myMin, prims[i]->getMinShiny());
    return myMin;
  }
  virtual SbVec3f  getMax() const {
    SbVec3f myMax(FLT_MIN, FLT_MIN, FLT_MIN);
    for (int i=0; i<prims.size(); i++)
      myMax = BBox::vecMax(myMax, prims[i]->getMax());
    return myMax;
  }
  virtual SbVec3f  getMaxShiny() const {
    SbVec3f myMax(FLT_MIN, FLT_MIN, FLT_MIN);
    for (int i=0; i<prims.size(); i++)
      myMax = BBox::vecMax(myMax, prims[i]->getMaxShiny());
    return myMax;
  }

};

class Instance : public Primitive {

private:
  Primitive *prim;
  SbMatrix m, mi;   // Matrix m and matrix m_inverse

public:
  Instance(Primitive *p, const SbMatrix &m_){
    prim = p;
    m = m_;
    mi = m.inverse();
  }
  ~Instance(){}

  void deletePrim(){ delete prim; }

  virtual PrimType getType() const { return PRIMTYPE_INSTANCE; }

  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);
  
  virtual SbVec3f  getMin() const {
    SbVec3f min;
    m.multVecMatrix(prim->getMin(), min);
    return min;
  }
  virtual SbVec3f  getMinShiny() const {
    SbVec3f min;
    m.multVecMatrix(prim->getMinShiny(), min);
    return min;
  }
  virtual SbVec3f  getMax() const {
    SbVec3f max;
    m.multVecMatrix(prim->getMax(), max);
    return max;
  }
  virtual SbVec3f  getMaxShiny() const {
    SbVec3f max;
    m.multVecMatrix(prim->getMaxShiny(), max);
    return max;
  }

  virtual Material* getMat() { return prim->getMat(); }
  virtual const Material* getMat() const { return prim->getMat(); }
  
  virtual SbVec3f getNormal(      const SbVec3f &bc) const{
    SbVec3f n;
    mi.transpose().multDirMatrix(prim->getNormal(bc), n);
    return n;
  }
  virtual SbVec3f getIntNormal(   const SbVec3f &bc) const {
    SbVec3f n;
    mi.transpose().multDirMatrix(prim->getIntNormal(bc), n);
    n.normalize();
    return n;
  }
  
  virtual SbColor getIntDiffColor(const SbVec3f &bc) const { return prim->getIntDiffColor(bc);}
  virtual SbColor getIntEmission( const SbVec3f &bc) const { return prim->getIntEmission(bc);}
  virtual bool    getDerivs(      const SbVec3f &bc, PrimDerivs &pDer) const{
    if (!prim->getDerivs(bc, pDer)) return false;
    SbMatrix tr = mi.transpose();
    tr.multDirMatrix(pDer.dPdU, pDer.dPdU);
    tr.multDirMatrix(pDer.dPdV, pDer.dPdV);
    tr.multDirMatrix(pDer.dNdU, pDer.dNdU);
    tr.multDirMatrix(pDer.dNdV, pDer.dNdV);
    return true;
  }
  
};

#endif  /* PRIMITIVE_H */
