
//'WWW''''''''WW''''WW'''''''WW''WWWW''WWW'''''''''''''''''''''WWW''''''''''''
//WW  W       WWW  WWW       WW  WW  W  WW                      WW            
// WW    WW   W WWW WW  WWW WWWW WW  W  WW  W WW  WWWW     WWW  WW   WWW  WWWW
//  WW  WW W  W  W  WW W WW  WW  WWWW   WW  W WW WWW      WW W  WW  W WW WWW  
//W  WW WW W  W     WW W WW  WW  WW     WW  W WW   WWW    WW    WW  W WW   WWW
// WWW   WW   W     WW  WWW  WW  WW    WWWW  WWW WWWW      WWW WWWW  WWW WWWW 
//                                                                            
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

// SoMatPlus is an Inventor node that you can place in
// your inventor scenes to add various properties to
// the standard SoMaterial node. These material properties
// will be used by pane in the appropriate way.
//
// SoMatPlus must come before a SoMaterial. Both
// will be stored in pane's Material class, which contains
// a copy of the SoMaterial info and a pointer to the
// SoMatPlus node.
//
// Example:
//
// SoMatPlus {
//   model schlick
//   rough1 1
//   rough2 1
//   isotropy1 1
//   isotropy2 1
//   R0 1
//   layers 1
//   ior1 1.0
//   ior2 1.5
//   atten .98 .99 .98
//   emisEcc 1
//   minDist .0001
// }
//
// Material {
//   diffuseColor .8 .8 .8
//   shininess 0
// }

#ifndef SOMATPLUSCLASS
#define SOMATPLUSCLASS

#include <iostream>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoSubNode.h>
#include <Inventor/fields/SoSFColor.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoSFEnum.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoSFVec2f.h>

using namespace std;

// like a MaterialPlus
class SoMatPlus : public SoNode {

  SO_NODE_HEADER(SoMatPlus);

  public:
  
  enum ReflModel {
    DEFAULT, SCHLICK, DATA, LIBRARY, BSSRDF, HIERSUBSURF, ASHIKHMIN
  };
  
  // fields
  
  SoSFEnum model; // model (schlick, data, library, ?)

  // Schlick
  // F0_r,g,b for top surface from Material diffuseColor
  // F0_r,g,b fro bottom surface from Material specularColor
  SoSFFloat rough1, rough2;  // roughness  (0=mirror, 1=diffuse (default))
  SoSFFloat isotropy1, isotropy2; // anisotropy (0=anisotropic,
                                  // 1=isotropic (symmetrical, default))
  SoSFFloat R0;              // reflection at normal incidence
                             // Default: 1
                             // -1 means auto-calculated based on index of refractions

  SoSFInt32 layers;            // 1 (default) or 2
  

  //SoSFFloat transparency; // 0 (opaque) - 1 (clear)
  SoSFFloat ior1;     // OUTSIDE index of refraction Default: 1.0 (air)
  SoSFFloat ior2;     // INSIDE index of refraction Default: 1.5 (glass)
  SoSFColor atten;    // attenuation (fraction) at 1m when going through inside
                      // Default (0.98, 0.99, 0.98)
                      // from specularColor ?


  
  SoSFFloat emisEcc;    // Light eccentricity (default 1 = diffuse)
  SoSFFloat minDist;  // minimum distance for intersection (default 0.0001)
  
  static void initClass();
  friend ostream& operator<< (ostream& os, const SoMatPlus& m);

  // constructor
  SoMatPlus();

  protected:

  virtual void GLRender(SoGLRenderAction *action);
  virtual void callback(SoCallbackAction *action);
  virtual void doAction(SoAction *action, const char *s);

  private:

  virtual ~SoMatPlus();

  //SoSFVec2f g, s, d;         // internal (private, automatically calculated)

  
  // Shader node
  
  // for colors
  //SoMaterial *mat;
  
  // Textures
  //const int maxTextures = 2;
  //Texture *text[maxTextures];
  
  // Material conversion
  // 1-shininess -> rough1, rough2
  // transparency:
  //    0: 0->transparency, 1->layers
  // != 0: transparency->transparency, 2->layers, -1->R0
  
  // Glass:
  //
  // diffuseColor 1 1 1   * R0 = ((n2-n1)/(n2+n1))^2 = 0.04 
  // sigma1 0
  // psi1 1
  //
  // ambientColor 1 1 1
  // sigma2 0
  // psi2 1
  //
  // layers 2
  // transparency 1
  // ior1 1.0
  // ior2 1.5
  // atten .98 .99 .98

  // Ideal diffuse
  //
  // diffuseColor 1 1 1   * R0 = 1.0
  // sigma1 1
  // psi1 1
  // layers 1
  // transparency 0

  // Ideal mirror
  //
  // diffuseColor 1 1 1 * R0 = 1
  // sigma1 0
  // psi1 1
  // layers 1
  // transparency 0
  
  // Green plastic
  //
  // diffuseColor 1 1 1  * R0 = .04 ?
  // sigma1 0.02
  // psi1 1
  //
  // ambientColor .1 .8 .1
  // sigma2 1
  // psi2 1
  //
  // layers 2
  // transparency 0

  // Smooth metal ("stainless" alloy)
  // diffuseColor 1 1 1 * R0 = 0.8
  // sigma1 0
  // psi1 1
  // layers 1
  // transparency 0
  
};

#endif /* SOMATPLUSCLASS */
